//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OdfDesignators.h --- Observation Data File designators related classes
//
//	DESCRIPTION:
//		Various classes are defined in conjunction with needed functionality
//		to access information in an ODF.
//
//	CHANGE HISTORY:
//		UL, 2001-12-06: 1.0.0: - initial version; partially extracted from
//								 Odf.h
//
//	AUTHOR:
//		U. Lammers, Thu Dec  6 16:05:50 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Observation Data File Interface Control Document, K. Galloway
//
//	$Header: /data/xmm02/cvsadmin/soc_repos/sas/packages/caloalutils/src/OdfDesignators.h,v 1.1 2001/12/06 17:20:52 ulammers Exp $
///

#ifndef ODF_DESIGNATORS_H
#define ODF_DESIGNATORS_H

#include <string>
#include <iosfwd>

#include "Xmm.h"

//
//	an OdfSet represents a data set with a name compliant with [2];
//	instantiation from a name; access methods to access the various
//	fields
//
class OdfSet {
public:
	OdfSet();
	OdfSet(const string &name);

	enum Type { TNONE = -1, EVENT_LIST, IMAGE, AUXILIARY, HOUSKEEPING,
				SPACECRAFT, SUMMARY };
	enum Format { FNONE = -1, COMPRESSED_FITS, COMPRESSED_ASCII, FITS, ASCII };

	const string		name() const { return _dirname + '/' + _basename; }
	const string		&dirname() const { return _dirname; }
	const string		&basename() const { return _basename; }
	unsigned			revolutionNr() const { return _revolutionNr; }
	const string		&proposalId() const { return _proposalId; }
	Xmm::Instrument		instrument() const { return _instr; }
	bool				scheduleFlagApplicable() const {
							return _scheduleFlagApplicable; }
	bool				scheduled() const { return _scheduleFlagApplicable &&
												   _scheduled; }
	bool				unscheduled() const { return _scheduleFlagApplicable &&
												   !_scheduled; }
	unsigned			exposureNr() const { return _exposureNr; }
	unsigned		    ccdNr() const { return _ccdNr; }
	unsigned			nodeNr() const { return _nodeNr; }
	Xmm::DataMode		dataMode() const { return _dataMode; }
	OdfSet::Type		type() const { return _type; }
	OdfSet::Format		format() const { return _format; }

	static bool			isValid(const std::string &name);
	bool				isValid() const { return _isValid; }

	~OdfSet() {}

private:
	//	data members
	bool				_isValid;
	string				_basename,
						_dirname;
	unsigned			_revolutionNr;
	string				_proposalId;
	Xmm::Instrument		_instr;
	bool				_scheduleFlagApplicable, _scheduled;
	unsigned			_exposureNr,
						_ccdNr,
						_nodeNr;
	Xmm::DataMode		_dataMode;
	OdfSet::Type		_type;
	OdfSet::Format		_format;
};

ostream &operator<<(ostream &os, const OdfSet &fname);

//
//	an identifier which uniquely identifies an instrument + exposure
//	in a given ODF; construction from 6-character string e.g. "R1S001";
//	methods to access the three fields
//
class InstrumentExposureIdentifier {
public:
	InstrumentExposureIdentifier(const std::string &id);
		//
		//	PURPOSE:
		//		Instantiate from 6-char string ("R1S001") or ODF ICD
		//		compliant set name
		//

	Xmm::Instrument		instrument() const { return _instr; }
	bool				scheduleFlagApplicable() const {
							return _scheduleFlagApplicable; }
	bool				scheduled() const { return _scheduleFlagApplicable &&
											_scheduled; }
	bool				unscheduled() const { return _scheduleFlagApplicable &&
											!_scheduled; }
	unsigned			exposureNumber() const { return _expNr; }

	std::string			exposureId() const;
	std::string			instrumentExposureId() const;

private:
	Xmm::Instrument		_instr;
	bool				_scheduleFlagApplicable, _scheduled;
	unsigned			_expNr;
};

#endif
