//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Xmm.c	--- implementation of class XMM
//
//	CHANGE HISTORY:
//		UL, 1997-??-??: 1.0.0: - first version
//		UL, 1998-08-27: 1.1.0: - changed PN CCD numbering scheme according
//								 to quadrantId * 3 + ccdId + 1
//		UL, 1999-02-23: 1.2.0: - new global object of type XmmPayLoad
//		UL, 1999-04-01: 1.3.0: - in-line with Xmm.h 2.3.0
//		UL, 1999-04-14: 1.4.0: - small change in mode detection logic for
//								 EMOS
//		UL, 1999-04-23: 1.4.1: - change in logic to detect EMOS mode W6
//								 (suggested by Jean Ballet)
//		UL, 1999-05-12: 1.4.2: - changed fatal -> error and changed all
//								 error/warning tags
//		UL, 1999-06-11: 1.4.3: - changed launch date to 1999-12-15T00:00:00
//							   - added orbital elements
//		UL, 1999-06-15: 1.4.4: - added implementation o
//								 EPNcam::toQuadrantColumnNumber()
//		UL, 1999-06-18: 1.4.5: - added another version of
//								 EPNcam::toQuadrantColumnNumber()
//		UL, 1999-06-24: 1.4.6: - added initialization routine for global
//								 xmmPayLoad object
//		UL, 1999-10-11: 1.5.0: - updated value of Xmm::missionStartTime
//		UL, 1999-10-20: 1.6.0: - added Xmm::emptyDataModeVec
//							   - added XmmInstrument::periodicHKmodes()
//		UL, 1999-11-17: 1.7.0: - brought logic in RGScam::toMode() in
//								 line with v7.0 [1] (changed SCITYPES)
//		UL, 2000-04-04: 1.8.0: - brought logic in RGScam::toMode() in
//								 line with v9.0 of [1] (changed SCITYPES)
//		UL, 2000-04-10: 1.9.0: - bug fix in EPNcam::toQuadrantColumnNumber()
//							   - brought in line with issue 2.4 of [2]
//		UL, 2000-05-24: 1.10.0:- EMOScam::toMode()/EPNcam::toMode() need
//								 now vector of CoreModeParameters as argument
//		UL, 2000-09-21: 1.11.0:- toMode() now returns pair<Mode, bool>
//		UL, 2000-09-29: 1.12.0:- toIntergralId() gives list of allowed
//								 values in error message when passed illegal
//								 string
//							   - various transparent changes to bring in
//								 line with changes in Xmm.h
//		UL, 2000-10-17: 1.13.0:- added XmmInstrument::scienceDataModes()
//		UL, 2000-11-07: 1.14.0:- change in mode detection logic for EMOS:
//								 outer CCDs are always operated in
//								 FullFrameMode mode; the mode detection logic
//								 cannot determine this from the mere contents
//								 of WINDOWX0/Y0/DX/DY + FRAMETIME alone
//								 because if the inner CCD is in RFS the
//								 frame time of the outer ones are extended
//								 to be an integer multiple of the inner
//								 FRAMETIME; this leads to ambiguities.
//		UL, 2001-01-10: 1.15.0:- added complete NASA/OIG TLE history covering
//								 period [1999-12-10, 2001-01-10]
//		UL, 2001-02-28: 2.17.0:- added couple of new classes:
//									* XmmGroundStation
//									* XmmMission
//		UL, 2001-03-22: 2.18.0:- added
//									* EMOScam::toHBR(Ccd)
//		UL, 2001-07-13: 2.19.0:- tweaked EPN mode detection logic such that
//								 ExtendedFullFrame is also recognized if
//								 FRAMETIME has a wrong value!
//		UL, 2001-07-16: 2.19.1:- associate DiscardedLinesData with Diagnostic
//								 mode
//							   - added ScienceEngineeringDataModes
//		UL, 2001-10-29: 2.20.0:- added 'using namespace std'
//		UL, 2002-03-07: 2.21.0:- EPNcam::toMode() now checks for F1294
//								 attribute to distinguish between FF and eFF
//								 (see also ODF-ICD DCP-140)
//		UL, 2002-04-16: 2.22.0:- updated GS propagation delays with latest
//								 data from Ed Serpell (should not have
//								 considerable impact on accuracy of time tags)
//
//	REFERENCES:
//		[1]: RGS Telemetry Structure, RGS-MSSL-SW-002, v9.0
//		[2]: ICD: Observation and Slew Data Files, K. Galloway
//
//  AUTHOR:
//		U. Lammers, Thu May 28 17:44:55 MET DST 1998
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/caloalutils/src/Xmm.cc,v 1.37 2002/04/16 12:33:23 ulammers Exp $
//

#include <algorithm>

#include "slatec++.h"
#include "Tracer.h"
#include "Xmm.h"
#include "Dal.h"
#include "DalUtils.h"
#include "StringPool.h"
#include "stringutil.h"

using namespace std;

static string _undef("UNDEFINED");

//
//	auxiliary function
//
namespace {

int toIntegralId(const StringPool &pool, const string &keyString,
				 const string &desc)
{
	if (keyString == _undef)
		return -1;

	string idstring(pool.expand(keyString));
	std::vector<string>::const_iterator sp = pool.begin();
	while (sp != pool.end()) {
		if ((*sp) == idstring)
			break;
		++sp;
	}

	if (sp != pool.end())
		return sp - pool.begin();
	else {
		errstr<<"`"<<keyString<<"' is not a valid "<<desc<< " identifier - "
				"the list of valid identifiers is:\n";
		unsigned N = pool.size();
		for(unsigned n=0; n<N; ++n)
			errstr<<pool[n]<<endl;
		errstr<<Sas::error("IllegalIdentifier");
	}

	return 0;
} // toIntegralId()


string dataType(const vector<const Attributable *> &av) {
	string dataType;
	try {
		dataType = uppercase(DalUtils::attribute(av, "DATATYPE")->asString());
	} catch (...) {
		errstr<<"Unable to find mandatory attribute `DATATYPE' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributeNotFound");
	}

	return dataType;
} // dataType()


vector<int> windowCoordinates(const vector<const Attributable *> &av)
{
	vector<int> v;
	try {
		v.push_back(DalUtils::attribute(av, "WINDOWX0")->asInt());
		v.push_back(DalUtils::attribute(av, "WINDOWY0")->asInt());
		v.push_back(DalUtils::attribute(av, "WINDOWDX")->asInt());
		v.push_back(DalUtils::attribute(av, "WINDOWDY")->asInt());
	} catch(...) {
		errstr<<"Unable to find manatory attributes `WINDOWX0', `WINDOWY0', "
				"`WINDOWDX', `WINDOWDY' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return v;
} // windowCoordinates()

pair<unsigned, string> frameTime(const vector<const Attributable *> &av)
{
	pair<unsigned, string> ft;
	try {
		const Attribute *frt = DalUtils::attribute(av, "FRMTIME");
		ft = make_pair(unsigned(frt->asInt()), frt->units());
	} catch(...) {
		errstr<<"Unable to find manatory attribute `FRMTIME' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return ft;
} // frameTime()

unsigned ccdMode(const vector<const Attributable *> &av)
{
	unsigned ccdm = 0;
	try {
		ccdm = DalUtils::attribute(av, "CCDMODE")->asInt();
	} catch(...) {
		errstr<<"Unable to find manatory attribute `CCDMODE' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return ccdm;
} // ccdMode()

unsigned sciType(const vector<const Attributable *> &av)
{
	unsigned st = 0;
	try {
		st = DalUtils::attribute(av, "SCITYPE")->asInt();
	} catch(...) {
		errstr<<"Unable to find manatory attribute `CCDMODE' in `"
			  <<av[0]->qualifiedName()<<"'"
			  <<Sas::error("MandatoryAttributesNotFound");
	}

	return st;
} // sciType()

//
//	access attribute F1294 ("FRAME_TIME_PARAMETER" of EPN) - possible
//	values [0, 3] or -1 (not present)
//
int frameTimeParameter(const vector<const Attributable *> &av)
{
	int ftp = -1;
	try {
		ftp = DalUtils::attribute(av, "F1294")->asInt();
	} catch (...) { }

	return ftp;
} // frameTimeParameter()

}


//
//	---( Xmm )--------------------------------------------------------------
//
const STime Xmm::missionReferenceTime("1997-12-31T23:58:56.816"); // UTC
const STime Xmm::missionStartTime("1999-12-10T14:32:00.000");
const STime Xmm::missionEndTime(Xmm::missionStartTime + 10. *
								365. * STime::secondsPerDay);

const std::vector<Node> Xmm::emptyNodeVec;
const std::vector<Mode> Xmm::emptyModeVec;
const std::vector<DataMode> Xmm::emptyDataModeVec;
const std::vector<Filter> Xmm::emptyFilterVec;

static const char *__ins[] = { INSTRUMENT_STRINGS, 0 };
static const char *__insShort[] = { INSTRUMENT_SHORT_STRINGS, 0 };
static const StringPool _instruments(__ins, StringPool::ABORT);
static const StringPool _instrumentsShort(__insShort, StringPool::ABORT);

Xmm::Instrument Xmm::toInstrument(const string &insString)
{
	return S_CAST(toIntegralId((insString.length() > 2 ? _instruments
													   : _instrumentsShort),
							   insString, "instrument"), Xmm::Instrument);
} // Xmm::toInstrumentId()

const string &Xmm::toString(Instrument instr, bool shortForm)
{
	if (instr == Xmm::INSTRUMENT_NONE)
		return _undef;

	return !shortForm ? _instruments[UNT(instr)]
					  : _instrumentsShort[UNT(instr)];
} // Xmm::toString

ostream &operator<<(ostream &os, Instrument instrument)
{
	os<<INT(instrument)<<" ["<<Xmm::toString(instrument)<<"]";
	return os;
} // operator<<


//
//	nodes
//
static const char *__nodes[] = { NODE_STRINGS, 0 };
static const StringPool _nodes(__nodes, StringPool::ABORT);

Xmm::Node Xmm::toNode(const string &nodeString)
{
	return S_CAST(toIntegralId(_nodes, nodeString, "node"), Xmm::Node);
} // Xmm::toNode()

Xmm::Node Xmm::toNode(unsigned node)
{
	if (node > N_NODES)
		errstr<<"Cannot convert "<<node<<" to Node id - must be in range "
				"[0, "<<N_NODES<<"]\n";

	return S_CAST(node, Xmm::Node);
} // Xmm::toNode()

const string &Xmm::toString(Node node)
{
	if (node == Xmm::NODE_NONE)
		return _undef;

	return _nodes[UNT(node)];
} // Xmm::toString()

ostream &operator<<(ostream &os, Node node)
{
	os<<INT(node)<<" ["<<Xmm::toString(node)<<"]";
	return os;
} // operator<<


//
//	data modes
//
static const char *__modes[] = { DATA_MODE_STRINGS, 0 };
static const StringPool _dataModes(__modes, StringPool::ABORT);

Xmm::DataMode Xmm::toDataMode(const string &modeString)
{
	return S_CAST(toIntegralId(_dataModes, modeString, "data mode"),
				  Xmm::DataMode);
} // Xmm::toDataMode()


const string &Xmm::toString(DataMode dataMode)
{
	if (dataMode == Xmm::DATA_MODE_NONE)
		return _undef;

	return _dataModes[dataMode];
} // Xmm::toString()

ostream &operator<<(ostream &os, DataMode mode)
{
	os<<INT(mode)<<" ["<<Xmm::toString(mode)<<"]";
	return os;
} // operator<<


//
//	camera modes
//
static const char *__camModes[] = { MODE_STRINGS, 0 };
static const StringPool _camModes(__camModes, StringPool::ABORT);

Xmm::Mode Xmm::toMode(const string &modeString)
{
	return S_CAST(toIntegralId(_camModes, modeString, "camera mode"),
				  Xmm::Mode);
} // Xmm::toMode()

Xmm::Mode Xmm::toMode(unsigned mode)
{
	if (mode > N_MODES)
		errstr<<"Cannot convert "<<mode<<" to Mode id - must be in range "
				"[0, "<<N_MODES<<"]\n";

	return S_CAST(mode, Xmm::Mode);
} // Xmm::toMode()

const string &Xmm::toString(Mode mode)
{
	if (mode == Xmm::MODE_NONE)
		return _undef;

	return _camModes[mode];
} // Xmm::toString()


ostream &operator<<(ostream &os, Mode mode)
{
	os<<INT(mode)<<" ["<<Xmm::toString(mode)<<"]";
	return os;
} // operator<<


//
//	filters
//
static const char *__filters[] = { FILTER_STRINGS, 0 };
static const StringPool _filters(__filters, StringPool::ABORT);

Xmm::Filter Xmm::toFilter(const string &filterString)
{
	return S_CAST(toIntegralId(_filters, filterString, "filter"),
				  Xmm::Filter);
} // Xmm::toFilter()

Xmm::Filter Xmm::toFilter(unsigned filter)
{
	if (filter > N_FILTERS)
		errstr<<"Cannot convert "<<filter<<" to filter id - must be in range "
				"[0, "<<N_FILTERS<<"]\n";

	return S_CAST(filter, Xmm::Filter);
} // Xmm::toFilter()

const string &Xmm::toString(Filter filter)
{
	if (filter == Xmm::FILTER_NONE)
		return _undef;

	return _filters[filter];
} // Xmm::toString()

ostream &operator<<(ostream &os, Filter filter)
{
	os<<INT(filter)<<" ["<<Xmm::toString(filter)<<"]";
	return os;
} // operator<<


//
//	---( XmmInstrument )---------------------------------------------------
//
static DataMode __periodicHKmodes[] = { Xmm::HK_P1, Xmm::HK_P2, Xmm::HK_P3,
									    Xmm::HK_P4, Xmm::HK_P5, Xmm::HK_P6,
									    Xmm::HK_P7, Xmm::HK_P8, Xmm::HK_P9 };
static std::vector<DataMode> _periodicHKmodes(
					__periodicHKmodes,
					__periodicHKmodes + NUM_OF(__periodicHKmodes));

const std::vector<DataMode> &XmmInstrument::periodicHKmodes() const
{
	return _periodicHKmodes;
} // XmmInstrument::periodicHKmodes


static std::vector<DataMode> _scienceDataModes;	//	no science modes
const std::vector<DataMode> &XmmInstrument::scienceDataModes() const
{
	return _scienceDataModes;
} // XmmInstrument::scienceDataModes


static std::vector<DataMode> _scienceEngineeringDataModes; // no modes

const std::vector<DataMode> &XmmInstrument::scienceEngineeringDataModes() const
{
	return _scienceEngineeringDataModes;
} // XmmInstrument::scienceEngineeringDataModes


unsigned XmmInstrument::modeIndex(Mode mode) const
{
	const std::vector<Mode> &modes = this->modes();
	unsigned N = modes.size();
	for (unsigned n=0; n<N; ++n)
		if (mode == modes[n])
			return n;

	errstr<<"`"<<Xmm::toString(mode)<<"' is not a valid mode of "
		  <<Xmm::toString(id())<<Sas::error("UnknownMode");

	return 0;	
} // XmmInstrument::modeIndex()


unsigned XmmInstrument::filterIndex(Filter filter) const
{
	const std::vector<Filter> &filters = this->filters();
	unsigned N = filters.size();
	for (unsigned n=0; n<N; ++n)
		if (filter == filters[n])
			return n;

	errstr<<"`"<<Xmm::toString(filter)<<"' is not a valid filter of "
		  <<Xmm::toString(id())<<Sas::error("UnknownFilter");

	return 0;	
} // XmmInstrument::filterIndex()


unsigned short XmmInstrument::toHBR(Ccd) const
{
	errstr<<"Generic XmmInstrument cannot convert CCD to HBR number"
		  <<Sas::error("UnsupportedOperation");

	return 0;
} // XmmInstrument::toHBR()


//
//	-----( EMOS )----------------------------------------------------------
//
//	nodes
//
static Node __EMOSnodes[] = { Xmm::PRIMARY_NODE,
							  Xmm::REDUNDANT_NODE };
const std::vector<Node> EMOScam::nodesVec(__EMOSnodes,
									 __EMOSnodes+NUM_OF(__EMOSnodes));


//
//	camera modes
//
static const Mode __EMOScamModes[] = {
	Xmm::PRIME_FULL_WINDOW, Xmm::PRIME_PARTIAL_RFS, Xmm::PRIME_PARTIAL_W2,
	Xmm::PRIME_PARTIAL_W3, Xmm::PRIME_PARTIAL_W4, Xmm::PRIME_PARTIAL_W5,
	Xmm::PRIME_PARTIAL_W6, Xmm::FAST_UNCOMPRESSED, Xmm::FAST_COMPRESSED,
	Xmm::OFFSET_VARIANCE, Xmm::CCD_DIAGNOSTIC, Xmm::EXTRA_HEATING_ANNEALING,
	Xmm::EXTRA_HEATING_DEICING, Xmm::EXTRA_HEATING_DECONTAM,
	Xmm::IN_FLIGHT_TEST };
const std::vector<Mode> EMOScam::modesVec(__EMOScamModes,
									 __EMOScamModes+NUM_OF(__EMOScamModes));

static const DataMode __EMOScamPeriodicHKmodes[] = {
	Xmm::PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> EMOScam::periodicHKmodesVec(
				__EMOScamPeriodicHKmodes,
				__EMOScamPeriodicHKmodes + NUM_OF(__EMOScamPeriodicHKmodes));

static const DataMode __EMOScamScienceDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::REDUCED_IMAGING, Xmm::COMPRESSED_TIMING
 };
const std::vector<DataMode> EMOScam::scienceDataModesVec(
				__EMOScamScienceDataModes,
				__EMOScamScienceDataModes + NUM_OF(__EMOScamScienceDataModes));

static const DataMode __EMOScamScienceEngineeringDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::REDUCED_IMAGING, Xmm::COMPRESSED_TIMING,
	Xmm::DATA_DIAGNOSTIC, Xmm::DATA_OFFSET_VARIANCE
 };
const std::vector<DataMode> EMOScam::scienceEngineeringDataModesVec(
				__EMOScamScienceEngineeringDataModes,
				__EMOScamScienceEngineeringDataModes +
								NUM_OF(__EMOScamScienceEngineeringDataModes));

//
//	figure the camera mode from datatype, x0, y0, dx, dy
//
static bool isCloseAbs(unsigned val, unsigned nominalVal, unsigned tolerance)
{
	return UNT(abs(INT(val - nominalVal))) <= tolerance;
} // isCloseAbs()

namespace {

Mode defaultEmosMode(const vector<const Attributable *> &av, bool silent)
{
	Mode mode = xmmPayLoad.epn.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];

	if (!silent)
		errstr<<"Cannot uniquely determine EMOS camera mode from tuple: "<<endl
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"   frameTime  = "<<frameTime(av).first<<" [ms]"<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INADEQUATE data "
				"which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultEmosMode()

}

pair<Mode, bool>
EMOScam::toMode(const vector<const Attributable *> &av,
				const std::vector<CoreModeParameters> &modeParameters,
				Ccd ccd, bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	unsigned dfModeOffset = UNT(xmmPayLoad.emos1.defaultMode());
	string dt = dataType(av);
	if (dt == "IMAGE.EL" || dt == "RIMAGE.EL") {
		//	get window coordinates
		vector<int> v = windowCoordinates(av);
		int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];

		//	frame time
		//	the following is ugly but unavoidable
		//
		pair<unsigned, string> ft = frameTime(av);
		if (ft.second.empty())
			//
			//	this is from the ODF in units of 100 ms
			//
			ft.first *= 100;
		else if (ft.second == "s" || ft.second == "sec")
			ft.first *= 1000;
		else if (ft.second != "ms")
			errstr<<"Unrecognized unit for attribute "
					"`FRMTIME\'"
				  <<Sas::error("UnrecognizedUnit");

		//
		//	image mode data
		//
		//	the mode is one of the following
		//
		//	ID MODE         X0,Y0,DX,DY        frameTime       EDUmode
		//	0  FULLWINDOW   0,0,600,600        2.6 ish         3
		//	1  PARTIAL RFS  0,0,600,600        0.2 ish         3
		//	2  W2           250,250,100,100    0.4 ish         3
		//	3  W3           150,150,300,300    0.9 ish         3
		//	4  W4           250,250,100,100    0.2 ish         3
		//	5  W5           150,150,300,300    0.7 ish         3
		//	6  W6           0,0,300,600        1.4 is          3
		//
		if (modeParameters.size() < 7) {
			errstr<<"Need at least mode parameters of EMOS camera modes:\n";
			for (unsigned n=0; n<7; ++n)
				errstr<<Xmm::toString(Xmm::toMode(n+dfModeOffset))<<endl;
			errstr<<"to determine effective istrument mode"
				  <<Sas::error("NoModeParameters");
		}
		std::vector<unsigned> modeIndices;
		for (unsigned n=0; n<modeParameters.size(); ++n)
			if (isCloseAbs(x0, modeParameters[n].windowX0, 5) &&
				isCloseAbs(y0, modeParameters[n].windowY0, 5) &&
				isCloseAbs(dx, modeParameters[n].windowDX, 10) &&
				isCloseAbs(dy, modeParameters[n].windowDY, 10) &&
				isCloseAbs(ft.first, modeParameters[n].frameTime, 100))
				//	found a match
				modeIndices.push_back(n);
		if (modeIndices.size() == 1)
			//	mode is unequivocally determined
			mode.first = Xmm::toMode(modeIndices[0] + dfModeOffset);
		else
			//	could not figure the mode - if it is an outer CCD the mode
			//	must be FullWindow
			if (ccd > 1)
				mode.first = Xmm::PRIME_FULL_WINDOW;
			else {
				if (modeIndices.size() > 1) {
					errstr<<"The camera mode is one of:\n";
					for (unsigned n=0; n<modeIndices.size(); ++n)
						errstr<<Xmm::toString(Xmm::toMode(modeIndices[n]
														+ dfModeOffset))<<endl;
				}

				mode.first = defaultEmosMode(av, silent);
				mode.second = false;
			}
	} else if (dt == "TIME.EL")
		mode.first = Xmm::FAST_UNCOMPRESSED;
	else if (dt == "CTIME.EL")
		mode.first = Xmm::FAST_COMPRESSED;
	else if (dt == "OFFVAR.EL")
		mode.first = Xmm::OFFSET_VARIANCE;
	else if (dt == "DIAG.IM")
		mode.first = Xmm::CCD_DIAGNOSTIC;
	//
	//	the list of modes we have not checked for at this point is:
	//
	//		EXTRA_HEATING_ANNEALING
	//		EXTRA_HEATING_DEICING
	//		EXTRA_HEATING_DECONTAM
	//		IN_FLIGHT_TEST
	//
	else {
		mode.first = defaultEmosMode(av, silent);
		mode.second = false;
	}

	return mode;
} // EMOScam::toMode()


//
//	filters
//
static const Filter __EMOSfilters[] = {
	Xmm::OPEN, Xmm::CLOSED, Xmm::THIN1, Xmm::THIN2, Xmm::MEDIUM, Xmm::THICK,
	Xmm::CAL_OPEN, Xmm::CAL_CLOSED, Xmm::CAL_THIN1, Xmm::CAL_THIN2,
	Xmm::CAL_MEDIUM, Xmm::CAL_THICK };
const std::vector<Filter> EMOScam::filtersVec(__EMOSfilters,
									     __EMOSfilters+NUM_OF(__EMOSfilters));

unsigned short EMOScam::toHBR(Ccd ccd) const
{
	static bool initialized = false;
	static map<Ccd, unsigned short> ccdHBRmap;
	if (!initialized) {
		//
		//	initialzie CCD-HBR map according to following translation table
		//	(see email D. Lumb, dated 2001-03-21):
		//	HBR =   1       CCDNR = 1
		//	        3               2
		//	        7               3
		//	        5               4
		//	        4               5
		//	        8               6
		//	        6               7
		//
		ccdHBRmap[1] = 1;
		ccdHBRmap[2] = 3;
		ccdHBRmap[3] = 7;
		ccdHBRmap[4] = 5;
		ccdHBRmap[5] = 4;
		ccdHBRmap[6] = 8;
		ccdHBRmap[7] = 6;

		initialized = true;
	}

	map<Ccd, unsigned short>::const_iterator i = ccdHBRmap.find(ccd);
	if (i == ccdHBRmap.end())
		errstr<<"There is no HBR corresponding to CCD #"<<ccd
			  <<Sas::error("NoSuchHBR");

	return i->second;
} // EMOScam::toHBR()


//
//	-----( EPN )----------------------------------------------------------
//
//
//	camera modes
//
static const Mode __EPNcamModes[] = {
	Xmm::PRIME_FULL_WINDOW, Xmm::PRIME_FULL_WINDOW_EXTENDED,
	Xmm::PRIME_LARGE_WINDOW, Xmm::PRIME_SMALL_WINDOW,
	Xmm::FAST_TIMING, Xmm::FAST_BURST, Xmm::PRIME_FULL_OFFSET,
	Xmm::PRIME_LARGE_OFFSET, Xmm::OFFSET, Xmm::NOISE, Xmm::DIAGNOSTIC,
	Xmm::EXTRA_HEATING_DEICING, Xmm::EXTRA_HEATING_DECONTAM,
	Xmm::IN_FLIGHT_TEST };
const std::vector<Mode> EPNcam::modesVec(__EPNcamModes,
								    __EPNcamModes+NUM_OF(__EPNcamModes));

namespace {

Mode defaultEpnMode(const vector<const Attributable *> &av, bool silent)
{
	Mode mode = xmmPayLoad.epn.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];
	if (!silent)
		errstr<<"Cannot uniquely determine EPN camera mode from tuple: "
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"   frameTime  = "<<frameTime(av).first<<" [ms]"<<endl
			  <<"   ccdMode    = "<<ccdMode(av)<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INAPPROPRIATE data"
				" which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultEpnMode()

}

pair<Mode, bool>
EPNcam::toMode(const vector<const Attributable *> &av,
			   const std::vector<CoreModeParameters> &modeParameters,
			   bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	string dt = dataType(av);
	if (dt == "IMAGE.EL") {
		pair<unsigned, string> ft = frameTime(av);
		unsigned ccdm = ccdMode(av);

		//
		//	image mode data
		//
		switch (ccdm) {
			case 1: //	this can be Xmm::PRIME_FULL_WINDOW_EXTENDED
					//	or Xmm::PRIME_FULL_WINDOW; decide through
					//	frame time
					if (modeParameters.size() < 2)
						errstr<<"Need at least mode parameters of EPN camera "
								"modes:\n"
							  <<Xmm::toString(Xmm::PRIME_FULL_WINDOW)
							  <<endl
							  <<Xmm::toString(Xmm::PRIME_FULL_WINDOW_EXTENDED)
							  <<"\nto determine effective instrument mode"
							  <<Sas::error("NoModeParameters");
					if (isCloseAbs(ft.first,
								   modeParameters[0].frameTime, 10))
						mode.first = Xmm::PRIME_FULL_WINDOW;
					else if (isCloseAbs(ft.first,
							 modeParameters[1].frameTime, 10))
						mode.first = Xmm::PRIME_FULL_WINDOW_EXTENDED;
					else {
						//	there was no close match in the frame time;
						//	this is probably PRIME_FULL_WINDOW_EXTENDED
						//	where PMS/ODS is known to fail to write the
						//	correct frame time into the FRAMETIME attribute;
						//	see if F1294 is present if yes and it is >0
						//	this is PRIME_FULL_WINDOW_EXTENDED
						int ftp = frameTimeParameter(av);
						if (ftp >= 0)
							mode.first = ftp > 0
											? Xmm::PRIME_FULL_WINDOW_EXTENDED
											: Xmm::PRIME_FULL_WINDOW;
						else {
							//	no F1294 attribute present - we give a warning
							//	and continue with
							//	PRIME_FULL_WINDOW_EXTENDED
							string effstr(
								Xmm::toString(
											Xmm::PRIME_FULL_WINDOW_EXTENDED));
							errstr<<"The window coordinates indicate this to "
									"be either `"
								  <<Xmm::toString(Xmm::PRIME_FULL_WINDOW)
								  <<"' or `"<<effstr
								  <<"' mode but the read frame time "
								  <<ft.first<<" [ms] does not match the "
									"nominal values ("
								  <<modeParameters[0].frameTime<<'/'
								  <<modeParameters[1].frameTime<<" [ms]) "
									"of either of those and the parameter "
									"F1293 (FRAME_TIME_PARAMETER) is not "
									"present - the frame time is likely "
									"to be the latter where the lack of an "
									"accurate FRAMETIME keyword for `"<<effstr
								  <<"' is a known ODF problem - will "
									"continue with `"<<effstr
								  <<"'. If you have created a summary file "
									"with odfingest you can check the value "
									"of FRAME_TIME_PARAMETER for this PN "
									"exposure. If it is not equal to 0 this "
									"is indeed an `"<<effstr<<"' exposure."
								  <<Sas::warning("EpnFrameTimeMismatch");
							mode.first = Xmm::PRIME_FULL_WINDOW_EXTENDED;
						}
					}
					break;
			case 2: mode.first = Xmm::PRIME_LARGE_WINDOW; break;
			case 3: mode.first = Xmm::PRIME_SMALL_WINDOW; break;
			default: break;
		}
	} else if (dt == "TIME.EL")
		mode.first = Xmm::FAST_TIMING;
	else if (dt == "BURST.EL")
		mode.first = Xmm::FAST_BURST;
	else if (dt == "DIAG.IM" || dt == "DISCLIN.IM")
		mode.first = Xmm::DIAGNOSTIC;
	else if (dt == "OFFSET.IM")
		mode.first = Xmm::OFFSET;
	else if (dt == "NOISE.IM")
		mode.first = Xmm::NOISE;
	else {
		mode.first = defaultEpnMode(av, silent);
		mode.second = false;
	}

	return mode;
} // EPNcam::toMode()	


static const DataMode __EPNcamPeriodicHKmodes[] = {
	Xmm::MAIN_PERIODIC_HOUSEKEEPING, Xmm::ADD_PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> EPNcam::periodicHKmodesVec(
					__EPNcamPeriodicHKmodes,
					__EPNcamPeriodicHKmodes + NUM_OF(__EPNcamPeriodicHKmodes));

static const DataMode __EPNcamScienceDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::BURST };
const std::vector<DataMode> EPNcam::scienceDataModesVec(
				__EPNcamScienceDataModes,
				__EPNcamScienceDataModes + NUM_OF(__EPNcamScienceDataModes));

static const DataMode __EPNcamScienceEngineeringDataModes[] = {
	Xmm::IMAGING, Xmm::TIMING, Xmm::BURST,
	Xmm::DATA_DIAGNOSTIC, Xmm::DATA_OFFSET_VARIANCE, Xmm::OFFSET_DATA,
	Xmm::NOISE_DATA, Xmm::DISCARDED_LINES_DATA
 };
const std::vector<DataMode> EPNcam::scienceEngineeringDataModesVec(
				__EPNcamScienceEngineeringDataModes,
				__EPNcamScienceEngineeringDataModes +
								NUM_OF(__EPNcamScienceEngineeringDataModes));

//
//	filters
//
static const Filter __EPNfilters[] = {
	Xmm::OPEN, Xmm::CLOSED, Xmm::THIN1, Xmm::THIN2,
	Xmm::MEDIUM, Xmm::THICK, Xmm::CAL_OPEN, Xmm::CAL_CLOSED,
	Xmm::CAL_THIN1, Xmm::CAL_THIN2, Xmm::CAL_MEDIUM,
	Xmm::CAL_THICK };
const std::vector<Filter> EPNcam::filtersVec(__EPNfilters,
									    __EPNfilters+NUM_OF(__EPNfilters));

unsigned EPNcam::toCcdId(unsigned ccd)
{
	if (ccd > 12)
		errstr<<"Illegal chip number - "<<ccd<<endl
			  <<Sas::error("IllegalChipNumber");

	return (ccd - 1)%3;
} // EPNcam::toCcdId()


unsigned EPNcam::toQuadrantId(unsigned ccd)
{
	if (ccd > 12)
		errstr<<"Illegal chip number - "<<ccd<<endl
			  <<Sas::error("IllegalChipNumber");

	return (ccd - 1)/3;
} // EPNcam::toQuadrantId()


unsigned EPNcam::toChipNr(unsigned ccdId, unsigned quadrantId)
{
	if (ccdId > 2)
		errstr<<"Illegal CCD number ("<<ccdId<<")"
			  <<Sas::error("IllegalCCDNumber");
	if (quadrantId > 3)
		errstr<<"Illegal quadrant number ("<<quadrantId<<")"
			  <<Sas::error("IllegalQuadrantNumber");

	return quadrantId * 3 + ccdId + 1;
} // EPNcam::toChipNr()

//
//	ccd		: [1, 12]
//	rawX	: [1, 64]
//	return	: quadrant column number [0, 191]
//
unsigned EPNcam::toQuadrantColumnNumber(Ccd ccd, unsigned rawX)
{
	return toQuadrantColumnNumber(toQuadrantId(ccd), toCcdId(ccd), rawX);
} // EPNcam::toQuadrantColumnNumber()


//
//	quadId: [0..3]
//	ccdId : [0..2]
//	rawX  : [1..64]
//	return: [0..191]
//
unsigned EPNcam::toQuadrantColumnNumber(unsigned quadId, unsigned ccdId,
									    unsigned rawX)
{
	if (!(quadId%2)) {	//	quadId == 0 || quadId == 2)
		return (2-ccdId)*64 + 64 - rawX;
	} else {
		return ccdId * 64 + 64 - rawX;
	}

	return 0;
} // EPNcam::toQuadrantColumnNumber()


//
//	-----( RGS )----------------------------------------------------------
//
//	nodes
//
static Node __RGSnodes[] = { Xmm::PRIMARY_NODE,
							 Xmm::REDUNDANT_NODE };
const std::vector<Node> RGScam::nodesVec(__RGSnodes,
								    __RGSnodes+NUM_OF(__RGSnodes));
static unsigned __RGSocbs[] = { 1, 2, 3, 4, 5 };
const std::vector<unsigned> RGScam::ocbs(__RGSocbs,
									__RGSocbs+NUM_OF(__RGSocbs));

//
//	camera modes
//
static const Mode __RGScamModes[] = {
	Xmm::SPECTROSCOPY, Xmm::HIGH_EVENT_RATE,
	Xmm::HIGH_EVENT_RATE_WITH_SES, Xmm::HIGH_EVENT_RATE_WITH_SER,
	Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD,
	Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD,
	Xmm::DIAGNOSTIC_1X1, Xmm::DIAGNOSTIC_2X2, Xmm::DIAGNOSTIC_3X3,
	Xmm::DIAGNOSTIC_4X4, Xmm::DIAGNOSTIC_5X5, Xmm::SPECTROSCOPY_TEST,
	Xmm::DIAGNOSTIC_TEST };
const std::vector<Mode> RGScam::modesVec(__RGScamModes,
								    __RGScamModes+NUM_OF(__RGScamModes));

namespace {

Mode defaultRgsMode(const vector<const Attributable *> &av,
				    unsigned ocbFactor, bool silent)
{
	Mode mode = xmmPayLoad.rgs1.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];
	if (!silent)
		errstr<<"Cannot uniquely determine RGS camera mode from tuple: "
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"   ocbFactor  = "<<ocbFactor<<endl
			  <<"   sciType    = "<<sciType(av)<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INAPPROPRIATE data "
				"which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultRgsMode()

}

pair<Mode, bool>
RGScam::toMode(const vector<const Attributable *> &av, unsigned ocbFactor,
			   bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	string dt = dataType(av);
	if (dt == "SPECT.EL") {
		//
		//	spectroscopy mode
		//
		unsigned scienceType = sciType(av);
		switch (scienceType) {
			case 1: mode.first = Xmm::SPECTROSCOPY; break;
			case 2: mode.first = Xmm::HIGH_EVENT_RATE; break;
			case 3: mode.first = Xmm::HIGH_EVENT_RATE_WITH_SER; break;
			case 4: mode.first = Xmm::HIGH_EVENT_RATE_WITH_SES; break;
			default:
				errstr<<"Illegal value of SCITYPE ("<<scienceType<<") in "
						"spectroscopy data mode"
					  <<Sas::error("IllegalAttributeValue"); break;
		}
	} else if (dt == "HTR.EL") {
		//
		//	this could be Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD or
		//	HIGH_TIME_RESOLUTION_MULTIPLE_CCD
		//	for the moment assume the first
		//
		unsigned scienceType = sciType(av);
		if (scienceType != 5)
			errstr<<"SCITYPE cannot be "<<scienceType
				  <<" in High Time Resolution data mode"
				  <<Sas::error("IllegalAttributeValue");
		mode.first = Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD;
	} else if (dt == "DIAG.IM")
		switch (ocbFactor) {
			case 1: mode.first = Xmm::DIAGNOSTIC_1X1; break;
			case 2: mode.first = Xmm::DIAGNOSTIC_2X2; break;
			case 3: mode.first = Xmm::DIAGNOSTIC_3X3; break;
			case 4: mode.first = Xmm::DIAGNOSTIC_4X4; break;
			case 5: mode.first = Xmm::DIAGNOSTIC_5X5; break;
			default: errstr<<"Illegal value of CCDOCB ("<<ocbFactor<<") for "
							 "diagnostic mode"
						   <<Sas::error("IllegalAttributeValue"); break;
		}
	else {
		mode.first = defaultRgsMode(av, ocbFactor, silent);
		mode.second = false;
	}

	return mode;
} // Rgscam::toMode()


static const DataMode __RGScamPeriodicHKmodes[] = {
	Xmm::FULL_PERIODIC_HOUSEKEEPING, Xmm::TEMP_PERIODIC_HOUSEKEEPING };
const std::vector<DataMode> RGScam::periodicHKmodesVec(
					__RGScamPeriodicHKmodes,
					__RGScamPeriodicHKmodes + NUM_OF(__RGScamPeriodicHKmodes));

static const DataMode __RGScamScienceDataModes[] = {
	Xmm::DATA_SPECTROSCOPY, Xmm::HIGH_TIME_RESOLUTION };
const std::vector<DataMode> RGScam::scienceDataModesVec(
				__RGScamScienceDataModes,
				__RGScamScienceDataModes + NUM_OF(__RGScamScienceDataModes));

static const DataMode __RGScamScienceEngineeringDataModes[] = {
	Xmm::DATA_SPECTROSCOPY, Xmm::HIGH_TIME_RESOLUTION,
	Xmm::DATA_DIAGNOSTIC
 };
const std::vector<DataMode> RGScam::scienceEngineeringDataModesVec(
				__RGScamScienceEngineeringDataModes,
				__RGScamScienceEngineeringDataModes +
								NUM_OF(__RGScamScienceEngineeringDataModes));

//
//	-----( OM )----------------------------------------------------------
//
//	nodes
//
static Node __OMnodes[] = { Xmm::PRIMARY_NODE,
							Xmm::REDUNDANT_NODE };
const std::vector<Node> OMcam::nodesVec(__OMnodes,
								   __OMnodes+NUM_OF(__OMnodes));

//
//	camera modes
//
static const Mode __OMcamModes[] = {
	Xmm::IMAGE, Xmm::FAST,
	Xmm::RAW_DATA_BLUE_DSP12, Xmm::RAW_DATA_BLUE_DSP1, Xmm::RAW_DATA_BLUE_DSP2,
	Xmm::CENTROIDING_DATA, Xmm::FULL_FRAME, Xmm::CENTROIDING_CONFIRMATION,
	Xmm::INTENSIFIER_CHARACTERISTICS, Xmm::FLATFIELDING };
const std::vector<Mode> OMcam::modesVec(__OMcamModes,
								   __OMcamModes+NUM_OF(__OMcamModes));

namespace {
Mode defaultOmMode(const vector<const Attributable *> &av, bool silent)
{
	Mode mode = xmmPayLoad.om.defaultMode();
	vector<int> v = windowCoordinates(av);
	int &x0 = v[0], &y0 = v[1], &dx = v[2], &dy = v[3];
	if (!silent)
		errstr<<"Cannot uniquely determine OM camera mode from tuple: "
			  <<"   dataType   = "<<dataType(av)<<endl
			  <<"   x0/y0      = "<<x0<<"/"<<y0<<endl
			  <<"   dx/dy      = "<<dx<<"/"<<dy<<endl
			  <<"will continue with "<<mode<<" as effective "
				"mode - Please note: As a result of a possibly incorrect mode "
				"setting the retrieval of mode-dependent calibration "
				"quantities via the Cal is likely to yield INAPPROPRIATE data "
				"which ultimately might lead to INCORRECT scientific results"
			  <<Sas::warning("UnidentifiedCameraMode");

	return mode;
} // defaultOmMode()

}

pair<Mode, bool>
OMcam::toMode(const vector<const Attributable *> &av, bool silent)
{
	pair<Mode, bool> mode(Xmm::MODE_NONE, true);
	string dt = dataType(av);
	if (dt == "IMG.IM")
		//
		//	image mode
		//
		mode.first = Xmm::IMAGE;
	else if (dt == "FST.EL")
		//
		//	fast mode
		//
		mode.first = Xmm::FAST;
	else {
		mode.first = defaultOmMode(av, silent);
		mode.second = false;
	}

	return mode;
} // OMcam::toMode()


static const DataMode __OMcamPeriodicHKmodes[] = {
	Xmm::PERIODIC_HOUSEKEEPING };	
const std::vector<DataMode> OMcam::periodicHKmodesVec(
					__OMcamPeriodicHKmodes,
					__OMcamPeriodicHKmodes + NUM_OF(__OMcamPeriodicHKmodes));

static const DataMode __OMcamScienceDataModes[] = {
	Xmm::IMAGING, Xmm::DATA_FAST };	
const std::vector<DataMode> OMcam::scienceDataModesVec(
					__OMcamScienceDataModes,
					__OMcamScienceDataModes + NUM_OF(__OMcamScienceDataModes));

static const DataMode __OMcamScienceEngineeringDataModes[] = {
	Xmm::IMAGING, Xmm::DATA_FAST,
	Xmm::DATA_DIAGNOSTIC, Xmm::ENGINEERING1, Xmm::ENGINEERING2,
	Xmm::ENGINEERING3, Xmm::ENGINEERING4, Xmm::ENGINEERING5, Xmm::ENGINEERING6,
	Xmm::ENGINEERING7, Xmm::ENGINEERING1
 };
const std::vector<DataMode> OMcam::scienceEngineeringDataModesVec(
				__OMcamScienceEngineeringDataModes,
				__OMcamScienceEngineeringDataModes +
								NUM_OF(__OMcamScienceEngineeringDataModes));

//
//	filters
//
static const Filter __OMfilters[] = {
	Xmm::BLOCKED, Xmm::UVW2, Xmm::UVM2, Xmm::UVW1, Xmm::U, Xmm::B, Xmm::V,
	Xmm::WHITE, Xmm::MAGNI, Xmm::GRISM1, Xmm::GRISM2, Xmm::BARRED_U };
const std::vector<Filter> OMcam::filtersVec(__OMfilters,
									   __OMfilters+NUM_OF(__OMfilters));


//
//	----( XmmPayLoad )-------------------------------------------------------
//
XmmInstrument XmmPayLoad::xmm;
EMOScam1 XmmPayLoad::emos1;
EMOScam2 XmmPayLoad::emos2;
ERMcam XmmPayLoad::erm;
EPNcam XmmPayLoad::epn;
RGScam1	XmmPayLoad::rgs1;
RGScam2 XmmPayLoad::rgs2;
OMcam XmmPayLoad::om;

const std::vector<XmmInstrument *> &XmmPayLoad::instruments()
{
	static std::vector<XmmInstrument *> _ins;
	static bool first = true;

	if (first) {
		first = false;

		_ins.push_back(&emos1);
		_ins.push_back(&emos2);
		_ins.push_back(&erm);
		_ins.push_back(&epn);
		_ins.push_back(&rgs1);
		_ins.push_back(&rgs2);
		_ins.push_back(&om);
	}	

	return _ins;
} // XmmPayLoad::instruments()


const XmmInstrument &XmmPayLoad::instrument(Xmm::Instrument inst)
{
	switch(inst) {
		case Xmm::EMOS1	: return XmmPayLoad::emos1; break;
		case Xmm::EMOS2	: return XmmPayLoad::emos2; break;
		case Xmm::EPN	: return XmmPayLoad::epn; break;
		case Xmm::ERM	: return XmmPayLoad::erm; break;
		case Xmm::RGS1	: return XmmPayLoad::rgs1; break;
		case Xmm::RGS2	: return XmmPayLoad::rgs2; break;
		case Xmm::OM	: return XmmPayLoad::om; break;
		case Xmm::SC	: return XmmPayLoad::xmm; break;
		default:
			errstr<<"Unknown instrument `"<<toString(inst)<<"'"
				  <<Sas::error("UnknownInstrument");
			break;
	}

	return XmmPayLoad::xmm;
} // XmmPayLoad::instrument()


XmmOrbitSimulator::XmmOrbitSimulator(const STime &time)
	: KeplerOrbitSimulator(Xmm::missionStartTime -
						   Xmm::missionReferenceTime)
													//	so, when we call the
													//	position/velocity
													//	method with event
													//	time tags, the
													//	launch date corresponds
													//	to a perigee pass
{
	static const unsigned N_PAR = 5;

	double orbitalElem[][N_PAR+1] = {
		//
		//	NORAD TLE from http://oig1.gsfc.nasa.gov:
		//
		//	TIME					[yy.ddddd]
		//	inclination				[deg]
		//	R.A. of ascending node	[deg]
		//	eccentricity			[0<e<1]
		//	argument of perigee		[deg]
		//	orbital period			[1/day]
		//
		{ 1999.+100./365.,
		  40., 240., 0.9, 50., .55 },
		{ 1999.+344.608/365.,
		  39.9041, 237.978, 0.8851996, 55.327, 0.55095466 },
		{ 1999.+344.6085/365.,
		  39.9041, 237.9777, 0.8851996, 55.327, 0.55095466 },
		{ 1999.+351.66667/365.,
		  38.931, 235.912, 0.794729, 56.927, 0.50147251 },
		{ 1999.+356.43244/365.,
		  38.7261, 235.16, 0.7940363, 57.6308, 0.50144297 },
		{ 1999.+356.50779/365.,
		  39.003, 235.3762, 0.7951865, 57.5044, 0.50151554 },
		{ 1999.+358.42732/365.,
		  38.7641, 234.9416, 0.7938691, 57.9159, 0.50143288 },
		{ 1999.+362.41508/365.,
		  38.6963, 234.6988, 0.7941813, 58.0903, 0.50143296 },
		{ 1999.+364.40884/365.,
		  38.7162, 234.4952, 0.7945439, 58.1824, 0.50142631 },
		{ 2000.+3.3970459/366.,
		  38.5478, 234.308, 0.795785, 58.0848, 0.50137751 },
		{ 2000.+9.5021638/366.,
		  38.488, 234.201, 0.7989415, 58.278, 0.50136759 },
		{ 2000.+15.375/366.,
		  38.3625, 233.7652, 0.7983233, 58.5975, 0.50128056 },
		{ 2000.+23.493937/366.,
		  38.5466, 233.3665, 0.7984616, 59.1144, 0.50150066 },
		{ 2000.+25.375/366.,
		  38.0587, 232.8394, 0.7993743, 59.2953, 0.50132631 },
		{ 2000.+25.502164/366.,
		  38.071, 232.828, 0.7994795, 59.341, 0.50130018 },
		{ 2000.+27.375/366.,
		  38.0311, 232.7572, 0.7994466, 59.3504, 0.50132179 },
		{ 2000.+31.502164/366.,
		  38.002, 232.518, 0.8002953, 59.58, 0.50136388 },
		{ 2000.+35.333333/366.,
		  37.8677, 232.2021, 0.8006114, 59.8586, 0.50141219 },
		{ 2000.+37.333333/366.,
		  37.7364, 232.0443, 0.8003598, 60.0142, 0.5013854 },
		{ 2000.+41.333333/366.,
		  37.3779, 231.8996, 0.7993935, 60.1279, 0.50147606 },
		{ 2000.+43.333333/366.,
		  37.3698, 231.8231, 0.7993432, 60.1697, 0.5012105 },
		{ 2000.+47.291667/366.,
		  37.5675, 231.3228, 0.8002899, 60.5313, 0.501277 },
		{ 2000.+51.291667/366.,
		  37.436, 230.8269, 0.7997824, 60.9363, 0.50117407 },
		{ 2000.+51.464656/366.,
		  37.442, 230.806, 0.799864, 60.982, 0.50131555 },
		{ 2000.+53.464656/366.,
		  37.429, 230.695, 0.7995385, 61.048, 0.50128101 },
		{ 2000.+55.291667/366.,
		  37.4239, 230.621, 0.7995749, 61.0849, 0.50130176 },
		{ 2000.+57.291667/366.,
		  37.4115, 230.5403, 0.7998324, 61.1404, 0.50134666 },
		{ 2000.+61.370912/366.,
		  37.348, 230.288, 0.8003056, 61.351, 0.5011945 },
		{ 2000.+63.25/366.,
		  37.3805, 230.0469, 0.800056, 61.552, 0.50121324 },
		{ 2000.+63.370912/366.,
		  37.304, 230.109, 0.800176, 61.512, 0.50119891 },
		{ 2000.+67.370912/366.,
		  37.243, 229.803, 0.7992313, 61.76, 0.50111691 },
		{ 2000.+69.25/366.,
		  37.2012, 229.7491, 0.7986935, 61.7478, 0.50102555 },
		{ 2000.+71.25/366.,
		  37.2156, 229.6837, 0.798731, 61.7588, 0.50101028 },
		{ 2000.+75.25/366.,
		  37.1209, 229.3574, 0.7994453, 61.995, 0.50117015 },
		{ 2000.+77.208333/366.,
		  37.106, 229.1269, 0.7991809, 62.194, 0.50111553 },
		{ 2000.+79.208333/366.,
		  37.0832, 228.9509, 0.7987042, 62.3325, 0.50101125 },
		{ 2000.+83.208333/366.,
		  37.1175, 228.8524, 0.7986812, 62.337, 0.50099197 },
		{ 2000.+85.166667/366.,
		  37.1104, 228.7923, 0.799026, 62.3705, 0.5010438 },
		{ 2000.+87.166667/366.,
		  37.0882, 228.6897, 0.7993412, 62.454, 0.50108616 },
		{ 2000.+89.208333/366.,
		  37.0048, 228.5708, 0.7994546, 62.5329, 0.50109173 },
		{ 2000.+91.208333/366.,
		  36.9746, 228.4041, 0.7992925, 62.6817, 0.50109528 },
		{ 2000.+93.208333/366.,
		  36.9642, 228.2633, 0.7989348, 62.7869, 0.50103752 },
		{ 2000.+95.321464/366.,
		  36.943, 228.186, 0.7986343, 62.845, 0.50095295 },
		{ 2000.+97.125/366.,
		  36.9867, 228.1404, 0.7983472, 62.8294, 0.50093285 },
		{ 2000.+99.125/366.,
		  36.9836, 228.1039, 0.7986819, 62.813, 0.50100094 },
		{ 2000.+99.166667/366.,
		  36.9559, 228.1129, 0.7986775, 62.7962, 0.5009989 },
		{ 2000.+101.125/366.,
		  36.9162, 227.9875, 0.7992767, 62.8721, 0.50109291 },
		{ 2000.+103.16667/366.,
		  36.8287, 227.7806, 0.7995101, 63.0333, 0.50110911 },
		{ 2000.+105.125/366.,
		  36.7886, 227.5942, 0.7992909, 63.1721, 0.50104045 },
		{ 2000.+107.16667/366.,
		  36.801, 227.4861, 0.7990461, 63.2259, 0.50097198 },
		{ 2000.+109.32146/366.,
		  36.81, 227.466, 0.7992524, 63.228, 0.50094705 },
		{ 2000.+111.125/366.,
		  36.8169, 227.4384, 0.7995471, 63.1872, 0.501006 },
		{ 2000.+113.125/366.,
		  36.8747, 227.3221, 0.8000756, 63.2708, 0.50106155 },
		{ 2000.+117.125/366.,
		  36.8421, 227.0794, 0.8007979, 63.4395, 0.50113684 },
		{ 2000.+117.32146/366.,
		  36.743, 227.131, 0.8008221, 63.474, 0.50106287 },
		{ 2000.+119.125/366.,
		  36.6435, 227.0575, 0.8005921, 63.5035, 0.50113735 },
		{ 2000.+123.125/366.,
		  36.6563, 226.8137, 0.8001154, 63.6762, 0.50104238 },
		{ 2000.+125.08333/366.,
		  36.6584, 226.7875, 0.8002517, 63.6526, 0.50102058 },
		{ 2000.+127.125/366.,
		  36.6456, 226.7194, 0.8008932, 63.6806, 0.50111396 },
		{ 2000.+127.32146/366.,
		  36.664, 226.702, 0.8011391, 63.712, 0.50093654 },
		{ 2000.+129.08333/366.,
		  36.5881, 226.5428, 0.801611, 63.8104, 0.50120803 },
		{ 2000.+131.08333/366.,
		  36.4558, 226.3234, 0.8017728, 63.9794, 0.5011835 },
		{ 2000.+133.04167/366.,
		  36.4243, 226.1599, 0.8016405, 64.0976, 0.50110655 },
		{ 2000.+135.08333/366.,
		  36.5088, 226.0798, 0.8017049, 64.1203, 0.50111023 },
		{ 2000.+137.08333/366.,
		  36.583, 226.0148, 0.8021499, 64.1095, 0.50114891 },
		{ 2000.+137.16718/366.,
		  36.495, 226.039, 0.8022281, 64.15, 0.50102737 },
		{ 2000.+140.29167/366.,
		  36.5596, 225.9128, 0.8030731, 64.1763, 0.50125836 },
		{ 2000.+141.04167/366.,
		  36.549, 225.868, 0.803397, 64.2141, 0.5012369 },
		{ 2000.+143.04167/366.,
		  36.5112, 225.7274, 0.803892, 64.3419, 0.50125792 },
		{ 2000.+145.04167/366.,
		  36.4176, 225.5841, 0.8040191, 64.4977, 0.50129116 },
		{ 2000.+147/366.,
		  36.3802, 225.4222, 0.804012, 64.6412, 0.50125749 },
		{ 2000.+149/366.,
		  36.3557, 225.2908, 0.8038582, 64.7452, 0.50120228 },
		{ 2000.+153.16718/366.,
		  36.301, 225.156, 0.8043142, 64.838, 0.5010127 },
		{ 2000.+155.04167/366.,
		  36.2616, 225.0582, 0.804945, 64.835, 0.50125411 },
		{ 2000.+157/366.,
		  36.1878, 224.8071, 0.8055942, 65.0432, 0.50135303 },
		{ 2000.+159/366.,
		  36.1382, 224.5319, 0.8056135, 65.3019, 0.50132379 },
		{ 2000.+160.95833/366.,
		  36.1106, 224.3704, 0.8055044, 65.4214, 0.50124966 },
		{ 2000.+162.95833/366.,
		  36.1039, 224.29, 0.8056549, 65.4657, 0.50123316 },
		{ 2000.+163/366.,
		  36.156, 224.3098, 0.8056539, 65.461, 0.50125662 },
		{ 2000.+164.95833/366.,
		  36.1462, 224.2444, 0.8061194, 65.499, 0.50129402 },
		{ 2000.+166.95833/366.,
		  36.1231, 224.1439, 0.8067451, 65.5829, 0.50134726 },
		{ 2000.+167/366.,
		  36.0897, 224.1157, 0.8067529, 65.5609, 0.50134555 },
		{ 2000.+168.95833/366.,
		  36.0514, 223.9667, 0.8073313, 65.6991, 0.50138897 },
		{ 2000.+170.83333/366.,
		  36.0038, 223.7822, 0.8077009, 65.8747, 0.50142131 },
		{ 2000.+172.95833/366.,
		  35.9578, 223.5719, 0.8078121, 66.0758, 0.50139353 },
		{ 2000.+174.95833/366.,
		  35.879, 223.3833, 0.8076102, 66.2991, 0.5014066 },
		{ 2000.+176.91667/366.,
		  35.8531, 223.2371, 0.807401, 66.4201, 0.50135031 },
		{ 2000.+178.91667/366.,
		  35.8385, 223.1386, 0.8073622, 66.4834, 0.50130776 },
		{ 2000.+180.95833/366.,
		  35.8385, 223.1023, 0.8073622, 66.5347, 0.50130776 },
		{ 2000.+182.91667/366.,
		  35.7634, 222.8644, 0.8084816, 66.646, 0.50150587 },
		{ 2000.+184.91667/366.,
		  35.7634, 222.8284, 0.8084816, 66.6969, 0.50150587 },
		{ 2000.+186.91667/366.,
		  35.6266, 222.2629, 0.8087164, 67.1651, 0.50145793 },
		{ 2000.+190.91667/366.,
		  35.5786, 222.0098, 0.8084962, 67.3948, 0.50141836 },
		{ 2000.+192.875/366.,
		  35.5616, 221.9094, 0.8089106, 67.4737, 0.5014666 },
		{ 2000.+194.91667/366.,
		  35.6211, 221.7588, 0.8095907, 67.5589, 0.50152834 },
		{ 2000.+196.875/366.,
		  35.579, 221.5669, 0.8099578, 67.7428, 0.50156255 },
		{ 2000.+199.87529/366.,
		  35.39, 221.295, 0.8202748, 67.394, 0.45803221 },
		{ 2000.+202.875/366.,
		  35.465, 220.9768, 0.8096681, 68.3105, 0.50155724 },
		{ 2000.+204.83333/366.,
		  35.3717, 220.8218, 0.8092837, 68.3777, 0.50151389 },
		{ 2000.+206.83333/366.,
		  35.3597, 220.7207, 0.8091956, 68.4439, 0.50149576 },
		{ 2000.+208.83333/366.,
		  35.3348, 220.6, 0.8095325, 68.5257, 0.50154203 },
		{ 2000.+209.125/366.,
		  35.3384, 220.5751, 0.8094202, 68.6072, 0.50157049 },
		{ 2000.+210.83333/366.,
		  35.2899, 220.3448, 0.8099094, 68.7997, 0.50171918 },
		{ 2000.+212.83333/366.,
		  35.22, 220.0078, 0.8099653, 69.0959, 0.50172086 },
		{ 2000.+214.83333/366.,
		  35.1811, 219.7413, 0.809543, 69.3192, 0.50162491 },
		{ 2000.+216.79167/366.,
		  35.1727, 219.6054, 0.8092469, 69.4144, 0.50157382 },
		{ 2000.+218.16667/366.,
		  35.1513, 219.5174, 0.8092448, 69.4373, 0.5017871 },
		{ 2000.+218.79167/366.,
		  35.1527, 219.4924, 0.8092717, 69.4539, 0.50161767 },
		{ 2000.+220.79167/366.,
		  35.1371, 219.3797, 0.8095826, 69.5464, 0.50165962 },
		{ 2000.+222.79167/366.,
		  35.1828, 219.2545, 0.8099713, 69.708, 0.50173311 },
		{ 2000.+224.75/366.,
		  35.1469, 219.053, 0.8101069, 69.9004, 0.50174706 },
		{ 2000.+226.75/366.,
		  35.1104, 218.8327, 0.8099836, 70.1114, 0.50172316 },
		{ 2000.+228.75/366.,
		  35.0824, 218.624, 0.809594, 70.3052, 0.501706 },
		{ 2000.+230.75/366.,
		  35.0261, 218.4344, 0.8090714, 70.4584, 0.50167039 },
		{ 2000.+232.75/366.,
		  35.0199, 218.327, 0.8086504, 70.5341, 0.50162703 },
		{ 2000.+234.75/366.,
		  35.0139, 218.2474, 0.8085702, 70.5759, 0.5016198 },
		{ 2000.+236.75/366.,
		  34.9945, 218.1226, 0.8088777, 70.6615, 0.50168805 },
		{ 2000.+238.70833/366.,
		  34.9455, 217.8719, 0.8091571, 70.872, 0.50177867 },
		{ 2000.+240.66667/366.,
		  34.8969, 217.5659, 0.8089376, 71.1347, 0.50173016 },
		{ 2000.+242.75/366.,
		  34.8641, 217.388, 0.8084395, 71.2666, 0.50164543 },
		{ 2000.+244.70833/366.,
		  34.8671, 217.3009, 0.8081433, 71.3112, 0.50160977 },
		{ 2000.+248.70833/366.,
		  34.8575, 217.1389, 0.8085144, 71.4186, 0.50170281 },
		{ 2000.+248.74755/366.,
		  34.848, 217.1, 0.8087315, 71.469, 0.50140772 },
		{ 2000.+250.70833/366.,
		  34.8355, 216.9899, 0.8087326, 71.5527, 0.50173239 },
		{ 2000.+252.70833/366.,
		  34.7781, 216.8053, 0.8086929, 71.7222, 0.5017059 },
		{ 2000.+254.66667/366.,
		  34.7537, 216.6164, 0.8084381, 71.8968, 0.50168686 },
		{ 2000.+256.66667/366.,
		  34.8099, 216.4163, 0.8080838, 72.0784, 0.50159142 },
		{ 2000.+260.625/366.,
		  34.8081, 216.2492, 0.8072219, 72.1783, 0.50150868 },
		{ 2000.+262.625/366.,
		  34.8044, 216.1959, 0.8072921, 72.1878, 0.50154244 },
		{ 2000.+266.66667/366.,
		  34.6513, 215.8326, 0.8076565, 72.4679, 0.50162117 },
		{ 2000.+270.625/366.,
		  34.6169, 215.4569, 0.8068803, 72.7522, 0.50145703 },
		{ 2000.+272.625/366.,
		  34.6492, 215.4829, 0.8066905, 72.768, 0.5013861 },
		{ 2000.+274.58333/366.,
		  34.6515, 215.4457, 0.8069463, 72.7714, 0.50143761 },
		{ 2000.+276.58333/366.,
		  34.6407, 215.359, 0.8073241, 72.8346, 0.50149475 },
		{ 2000.+278.58333/366.,
		  34.621, 215.2225, 0.8075973, 72.954, 0.50151407 },
		{ 2000.+280.625/366.,
		  34.5933, 215.0923, 0.8078816, 72.9898, 0.5014752 },
		{ 2000.+282.58333/366.,
		  34.5748, 214.933, 0.8076674, 73.1307, 0.50145187 },
		{ 2000.+284.58333/366.,
		  34.5656, 214.8071, 0.8072842, 73.2311, 0.50139874 },
		{ 2000.+288.58333/366.,
		  34.6472, 214.5744, 0.8067898, 73.355, 0.50129388 },
		{ 2000.+290.54167/366.,
		  34.6373, 214.52, 0.8071448, 73.3636, 0.50136838 },
		{ 2000.+292.58333/366.,
		  34.5496, 214.4625, 0.8077382, 73.3959, 0.50142961 },
		{ 2000.+294.54167/366.,
		  34.5096, 214.2147, 0.8078791, 73.5997, 0.50143857 },
		{ 2000.+296.54167/366.,
		  34.4905, 214.0166, 0.8076176, 73.7516, 0.50136115 },
		{ 2000.+300.54167/366.,
		  34.5103, 213.8481, 0.8076713, 73.8101, 0.50127501 },
		{ 2000.+302.5/366.,
		  34.5053, 213.8018, 0.8081617, 73.8259, 0.50134033 },
		{ 2000.+304.5/366.,
		  34.4878, 213.6971, 0.8087108, 73.9095, 0.50137802 },
		{ 2000.+308.54167/366.,
		  34.4824, 213.3497, 0.8095186, 74.1258, 0.50137581 },
		{ 2000.+310.5/366.,
		  34.4626, 213.1905, 0.8094165, 74.2658, 0.50133984 },
		{ 2000.+312.5/366.,
		  34.452, 213.0705, 0.8091897, 74.3595, 0.50129275 },
		{ 2000.+314.45833/366.,
		  34.3994, 213.0398, 0.8087724, 74.4714, 0.50121284 },
		{ 2000.+316.45833/366.,
		  34.3938, 212.9993, 0.8089893, 74.4687, 0.50120991 },
		{ 2000.+318.45833/366.,
		  34.3708, 212.8947, 0.8096084, 74.5287, 0.50130308 },
		{ 2000.+320.5/366.,
		  34.3033, 212.6322, 0.8101934, 74.7303, 0.50134918 },
		{ 2000.+322.45833/366.,
		  34.2636, 212.3557, 0.8102921, 74.9635, 0.5013197 },
		{ 2000.+322.64243/366.,
		  34.3, 212.364, 0.8104332, 74.99, 0.50136131 },
		{ 2000.+324.45833/366.,
		  34.2487, 212.1743, 0.8101303, 75.1014, 0.50123609 },
		{ 2000.+328.45833/366.,
		  34.2639, 212.0338, 0.8106279, 75.1645, 0.50122631 },
		{ 2000.+330.41667/366.,
		  34.2477, 211.9361, 0.8112665, 75.2389, 0.50127425 },
		{ 2000.+332.41667/366.,
		  34.2221, 211.7757, 0.8118603, 75.3846, 0.50133319 },
		{ 2000.+334.45833/366.,
		  34.3295, 211.4427, 0.8125538, 75.7571, 0.50132984 },
		{ 2000.+336.41667/366.,
		  34.301, 211.2315, 0.8126706, 75.955, 0.50131558 },
		{ 2000.+338.41667/366.,
		  34.2797, 211.0422, 0.812574, 76.1273, 0.50127782 },
		{ 2000.+340.41667/366.,
		  34.2648, 210.8912, 0.8123993, 76.2499, 0.50127102 },
		{ 2000.+342.41667/366.,
		  34.1011, 210.951, 0.8120196, 76.097, 0.50118687 },
		{ 2000.+344.375/366.,
		  34.0861, 210.8617, 0.8123692, 76.1468, 0.5012132 },
		{ 2000.+346.375/366.,
		  34.0511, 210.6755, 0.8130479, 76.2913, 0.50130809 },
		{ 2000.+348.41667/366.,
		  34.0016, 210.3506, 0.8135125, 76.5749, 0.50135437 },
		{ 2000.+350.375/366.,
		  33.9673, 210.0366, 0.8134366, 76.8469, 0.50128053 },
		{ 2000.+352.375/366.,
		  33.9562, 209.8518, 0.8132603, 76.9915, 0.50120463 },
		{ 2000.+354.375/366.,
		  33.9513, 209.749, 0.8134047, 77.0599, 0.50120114 },
		{ 2000.+356.375/366.,
		  33.939, 209.6388, 0.813868, 77.1451, 0.50124514 },
		{ 2000.+358.33333/366.,
		  33.9654, 209.4126, 0.8150084, 77.4668, 0.50114275 },
		{ 2000.+360.375/366.,
		  33.9368, 209.1885, 0.8154493, 77.681, 0.50117474 },
		{ 2000.+362.33333/366.,
		  33.9076, 208.9349, 0.815637, 77.9258, 0.50117563 },
		{ 2000.+364.25985/366.,
		  33.883, 208.6872, 0.8155563, 78.1627, 0.50114789 },
		{ 2001.+2.375/365.,
		  33.8384, 208.3574, 0.8146676, 78.3695, 0.50089489 },
		{ 2001.+4.3333333/365.,
		  33.8289, 208.24, 0.8145958, 78.4574, 0.5008785 },
	};

	unsigned nPoints = NUM_OF(orbitalElem);
	for (unsigned n=0; n<nPoints; ++n) {
		//	convert time to MET
		orbitalElem[n][0] = STime(orbitalElem[n][0], STime::DYN) -
								Xmm::missionReferenceTime;
		//	convert degrees to rad
		orbitalElem[n][1] *= DEG_TO_RAD;
		orbitalElem[n][2] *= DEG_TO_RAD;
		orbitalElem[n][4] *= DEG_TO_RAD;

		//	convert orbital period [rev/d] -> [s/rev]
		orbitalElem[n][5] = 86400./orbitalElem[n][5];
	}

	double t = time - Xmm::missionReferenceTime;	//	time at which
													//	orbital elements
													//	are sought
	double parValues[N_PAR];
	for (unsigned n=0; n<N_PAR; ++n) {
		std::vector<double> x(nPoints),
					   y(nPoints);
		for (unsigned i=0; i<nPoints; ++i) {
			x[i] = orbitalElem[i][0];
			y[i] = orbitalElem[i][n+1];
		}
		Slatec::CubicHermiteInterpolant<double> f(x, y);
		parValues[n] = f(t);
		//cout<<n<<' '<<parValues[n]<<endl;
	}

	OrbitalElements orb = { parValues[4],	//	period
							114000.,		//	apogee [km]
							7000.,			//	perigee [km]
							parValues[2],	//	eccentricity
							parValues[0],	//	orbit inclination [rad]
							parValues[1],	//	R.A. of ascending node [rad]
							parValues[3],	//	argument of perigee [rad]
						  };

	orbitalElements(orb);
} // XmmOrbitSimulator::XmmOrbitSimulator()


XmmGroundStation::XmmGroundStation(const EarthObject &where,
								   unsigned short id,
								   double timingDelay)
	: EarthPositionLocator(where, Xmm::missionReferenceTime),
	  _id(id), _timingDelay(timingDelay)
{
} // XmmGroundStation::XmmGroundStation()

//
//	-------------------------( XMM )------------------------------------------
//
const XmmPayLoad &XmmMission::payLoad()
{
	static XmmPayLoad *xmmPayLoadObj = new XmmPayLoad;

	return *xmmPayLoadObj;
} // XmmMission::payLoad()


const XmmDataHandler &XmmMission::dataHandler()
{
	static XmmDataHandler *xmmDataHandler = new XmmDataHandler;

	return *xmmDataHandler;
} // XmmMission::dataHandler()


const XmmOrbitSimulator &XmmMission::orbitSimulator()
{
	static XmmOrbitSimulator *xmmOrbitSimulator = new XmmOrbitSimulator;

	return *xmmOrbitSimulator;
} // XmmMission::orbitSimulator()


XmmMission::GroundStations XmmMission::_groundStations;


void XmmMission::checkGSset()
{
	if (!_groundStations.size()) {
		//
		//	following data taken from ESOC file Station.dat and data provided
		//	by Ed Serpell on 2001-02-28; the timing delay numbers are
		//	used as of XMCS v72.
		//
		XmmGroundStation
			Vilspa1(EarthObject(356.04841574 * DEG_TO_RAD,
								40.44256248 * DEG_TO_RAD,EarthObject::GEODETIC,
								656.455,
								"Villafranca-1"), 5, 474.17e-6),
												//	Vilspa2 value in absence
												//	of more specific data
			Vilspa2(EarthObject(356.04740072 * DEG_TO_RAD,
								40.44559434 * DEG_TO_RAD,EarthObject::GEODETIC,
								664.432,
								"Villafranca-2"), 13, 474.17e-6),
			Vilspa3(EarthObject(356.04841847 * DEG_TO_RAD,
								40.44333913 * DEG_TO_RAD,EarthObject::GEODETIC,
								650.833,
								"Villafranca-3"), 14, 474.17e-6),
												//	Vilspa2 value in absence
												//	of more specific data
			Perth(	EarthObject(115.88515805 * DEG_TO_RAD,
								-31.80252200 *DEG_TO_RAD,EarthObject::GEODETIC,
								22.160,
								"Perth"), 22, 465.54e-6),
			Kourou( EarthObject(307.19533724 * DEG_TO_RAD,
								5.25143810 * DEG_TO_RAD, EarthObject::GEODETIC,
								-14.558,
								"Kourou"), 21, 474.87e-6),
			Santiago(EarthObject(289.33363100 * DEG_TO_RAD,
								-33.15099900 *DEG_TO_RAD,EarthObject::GEODETIC,
								729.0,
								"Santiago"), 41, 597.00e-6 + 110.e-6);
													//	additional 110us yield
													//	better data alignment

		_groundStations.push_back(Vilspa1);
		_groundStations.push_back(Vilspa2);
		_groundStations.push_back(Vilspa3);
		_groundStations.push_back(Perth);
		_groundStations.push_back(Kourou);
		_groundStations.push_back(Santiago);
	}
} // XmmMission::checkGSset();


const XmmGroundStation &XmmMission::groundStation(
										XmmMission::GroundStationPlaces place)
{
	checkGSset();
	return _groundStations[UNT(place)];
} // XmmMission::groundStation()


class GSeq : public unary_function<XmmGroundStation, bool> {
	unsigned short _id;
public:
	explicit GSeq(unsigned short id) : _id(id) {}
	bool operator()(const XmmGroundStation &gs) { return gs.id()==_id; }
};


const XmmGroundStation &XmmMission::groundStation(unsigned short GSID)
{
	checkGSset();
	GroundStations::iterator i =
		find_if(_groundStations.begin(), _groundStations.end(), GSeq(GSID));
	if (i == _groundStations.end())
		errstr<<"There is no XMM ground station with ID "<<GSID
			  <<Sas::error("NoSuchGroundStation");

	return *i;
} // XmmMission::groundStation()
