//
//	test the cal
//
#include <iostream>
#include <iomanip>
#include <cmath>
#include "common.h"
#include "Celestial.h"
#include "CalOalUtility.h"
#include "EarthPositionLocator.h"

const double STEP = 45.;

using namespace std;

namespace {

static int scale(double v) { return static_cast<int>(
							v < 0. ? std::ceil(v - .5) : std::floor(v + .5)); }

int nearestInt(double x)
{
	return INT(x<0. ? ceil(x-.5) : floor(x+.5));
}

void checkAngle(double rad, int hundredthArcsec)
{
	assert(nearestInt(rad*RAD_TO_DEG*3600.*100.) == hundredthArcsec);
}

}

int main()
{
	using namespace CalOalUtility;

	assert(SQR(2) == 4);
	assert(CUB(2.) == 8.);

	checkAngle(Declination("-0:00:00.00").angle(), 0);
	checkAngle(Declination("-0:30:00.00").angle(), -180000);
	checkAngle(Declination("0:30:00.00").angle(), 180000);
	checkAngle(Declination("-0:59:59.99").angle(), -359999);
	checkAngle(Declination("0:59:59.99").angle(), 359999);
	checkAngle(Declination("-1:00:00.00").angle(), -360000);
	checkAngle(Declination("180.").angle(), 64800000);
	checkAngle(Declination("-180.").angle(), -64800000);
	checkAngle(RightAscension("12.").angle(), 64800000);
	checkAngle(RightAscension("-12.").angle(), -64800000);

	for (int iphi = 0; iphi<=8; ++iphi)
		for (int itheta = -2; itheta<=2; ++itheta)
			for (int ipsi = 0; ipsi<=8; ++ipsi) {
				double phi = STEP * iphi;
				double theta = STEP * itheta;
				double psi = STEP * ipsi;

				AttitudeMatrix a321(phi * DEG_TO_RAD,
									theta * DEG_TO_RAD,
									psi * DEG_TO_RAD,
									EulerAngles::Euler321);

				EulerAngles e = a321.toEulerAngles(EulerAngles::Euler321);
//std::cout<<scale(phi)<<" "<<scale(e.phi() * RAD_TO_DEG)<<std::endl;
//std::cout<<scale(theta)<<" "<<scale(e.theta() * RAD_TO_DEG)<<std::endl;
//std::cout<<scale(psi)<<" "<<scale(e.psi() * RAD_TO_DEG)<<std::endl;
				assert(scale(phi) == scale(e.phi() * RAD_TO_DEG) &&
					   scale(theta) == scale(e.theta() * RAD_TO_DEG) &&
					   scale(psi) == scale(e.psi() * RAD_TO_DEG));
			}

	//
	//	Ra/Dec test
	//
	double d1 = 105.;
	std::string d1s("07h00m00.00s");
	assert(std::fabs(RAD_TO_DEG*RightAscension("7").angle()/d1-1.)
							<1.e-8);

	double d2 = 10.*15. + .25*23. + 34.99/240.;
	std::string d2s("10h23m34.99s");
	assert(std::fabs(RAD_TO_DEG*RightAscension(d2s).angle()/d2-1.)<1.e-8);

	double d3 = -89.;
	std::string d3s("-89:00:00.00");
	assert(std::fabs(RAD_TO_DEG*Declination("-89").angle()/d3-1.)<1.e-8);

	double d4 = 17. + 12./60. + 34.76/3600.;
	std::string d4s("17:12:34.76");
	assert(std::fabs(RAD_TO_DEG*Declination(d4s).angle()/d4-1.)<1.e-8);

	assert(RightAscension(d1 * DEG_TO_RAD).toString() == d1s);
	assert(RightAscension(d2 * DEG_TO_RAD).toString() == d2s);
	assert(Declination(d3 * DEG_TO_RAD).toString(':') == d3s);
	assert(Declination(d4 * DEG_TO_RAD).toString(':') == d4s);

	EquatorialDirection s(RightAscension("0."), Declination(0.1));
	std::cout<<s<<std::endl;

	//
	//	test EarthPositionLocator
	//
	EarthObject Leiden(+4.485*DEG_TO_RAD, 52.155*DEG_TO_RAD,
					   EarthObject::GEODETIC, 12.);
	STime tref("2000-01-01");
	EarthPositionLocator epos(Leiden);
	StateVector v(epos.stateVector(0.));
	double sidT = RAD_TO_DEG*tref.siderealTime(+4.485*DEG_TO_RAD);
	std::cout<<"State vector: "<<v<<std::endl;
	std::cout<<"Sidereal time: "<<sidT<<" degrees"<<std::endl<<std::endl;
	assert((RAD_TO_DEG*atan2(v.position.y(), v.position.x()))/sidT-1.<1.e-10);

	//
	//	test the WCS stuff - see Greisen, Calabretta, Sect. 6
	//
	double deltaX = -.00277777 * DEG_TO_RAD, // [deg] - 10"
		   deltaY = -deltaX,
		   refX   = 256.,
		   refY   = 257.,
		   alphaP = 45.83, // [deg] RA at ref point
		   deltaP = 63.57; // [deg] Dec at ref point

	using namespace WCS;
	EquatorialDirection dir(RightAscension(alphaP * DEG_TO_RAD),
							Declination(deltaP * DEG_TO_RAD));
	TangentialGnomonicProjection wcsfix(dir,
										refX, deltaX, refY, deltaY);

	double testX[] = { refX, .5, 512.5 },
		   testY[] = { refY, .5, 512.5 };
	for (unsigned n=0; n<NUM_OF(testX); ++n) {
		ZenithalProjection::NativePlaneCoordinates p0(testX[n], testY[n]);
		EquatorialDirection d(wcsfix.toStandardSystem(p0));
		ZenithalProjection::NativePlaneCoordinates p1 =
													wcsfix.toNativeSystem(d);

		std::cout<<p0.first<<'/'<<p0.second<<" = "
				 <<std::setprecision(12)<<RAD_TO_DEG*d.ra().angle()<<'/'
				 <<RAD_TO_DEG*d.dec().angle()
				 <<" = "<<p1.first<<'/'<<p1.second<<std::endl;
	}

	EquatorialDirection vernalEquinox(RightAscension(0.), Declination(0.)),
						ed1(vernalEquinox), ed2(vernalEquinox);
	std::cout<<ed1.toApparentDirection(CartesianVector(1., 1., 0.), 1.)
			 <<std::endl;
	assert(std::fabs(ed1.ra().angle()-3.*M_HALF_PI)<1.e-9);

	std::cout<<ed2.toApparentDirection(CartesianVector(2., 0., 0.), 1.)
			 <<std::endl;
	assert(std::fabs(ed2.ra().angle()-M_PI)<1.e-9);

	Plane p(CartesianVector(0., 0., 0.), 
			CartesianVector(100., 0., 0.),
			CartesianVector(100., 50., 0.));
	std::cout<<p.normal()<<endl;
	assert((p.normal()-CartesianVector(0., 0., 1.)).norm() < 1.e-8);

	return 0;
}
